"""Environment management module.

This module allows 3D environment generation from templates, data structuring
and display in an easy way.

Modules:
    reader: track template reader module.
    track: track structuring and display module.

Author:
    Paulo Sanchez (@erlete)
"""
