"""Core functionalities module.

This module contains all classes required to provide with the core
functionality, including vector operations and display, color gradient
generation, etc.

Modules:
    color: color utilities module.
    vector: vector utilities module.

Author:
    Paulo Sanchez (@erlete)
"""
