"""Geometry module.

This module contains several classes that make it possible to display specific
surfaces in a 3D environment, such as track ring and drone models.

Modules:
    drone: drone model display module.
    ring: track ring display module.

Author:
    Paulo Sanchez (@erlete)
"""
