"""Simple Drone Control (SDC) package.

This package contains several libraries that allow the user to control a drone
in a simulated 3D environment. The model is purely mathematical and does not
consider certain physical or dynamic aspects of the real world, since it is
meant to be used as a learning tool.

Modules:
    api: API modules for simulation, tracks and drones.
    core: core modules such as vector and color utilities.
    environment: environment management and display module.
    geometry: geometry generation and display module.

Author:
    Paulo Sanchez (@erlete)
"""
